/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner and Hoifung Poon.
 * 
 * Copyright [2004-06] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef UNITPROPAGATION_H_
#define UNITPROPAGATION_H_

#include "inference.h"

class UnitPropagation : public Inference
{
 public:
  
  /**
   * There is nothing to do in the constructor for unit propagation.
   */
  UnitPropagation(VariableState* state, long int seed)
    : Inference(state, seed) {}

  ~UnitPropagation() {}
  
  /**
   * There is nothing to initialize in unit propagation.
   */
  void init() { return; }
  
  /**
   * Perform unit propagation on the clauses in the state.
   */
  void infer()
  {
    bool done = false;
      // We are done when no more propagation occurs
    while (!done)
    {
      done = true;
      for (int i = 0; i < state->getNumClauses(); i++)
      {
          // Simplify clause i using the atoms already fixed
        Array<int>* atoms = state->simplifyClauseFromFixedAtoms(i);
        if (atoms->size() == 0)
        {
          delete atoms;
          continue;
        }
        
          // If unit clause or negative weight, then fix the atom(s)
        long double cost = state->getClauseCost(i);
        if (atoms->size() == 1 || cost < 0)
        {
            // We are fixing atoms, so we have possible propagation
          done = false;
            // Fix each atom in the clause
          for (int j = 0; j < atoms->size(); j++)
          {
            int lit = (*atoms)[j];
              // Set value to true if pos. lit. and pos. clause
              // or neg. lit. and neg. clause
            bool value = ((lit > 0 && cost > 0) || (lit < 0 && cost < 0))
              ? true : false;
            state->fixAtom(abs(lit), value);
          }
        }
        delete atoms;
      }
    }
      // Done with UP, save this state
    state->saveLowState();
  }
  
 private:
 
   
};


#endif /*UNITPROPAGATION_H_*/
